"""Provides a Hive-partitioned, file-based I/O handler for Parquet format.

This handler stores datasets in a wide format (series as columns) with embedded metadata,
using a defined Hive-partitioned directory structure. For example:

.. code-block::

    <repository_root>/
    ├── data_type=AS_OF_AT/
    │   └── dataset=my_versioned_dataset/
    │       ├── as_of=2023-01-01T120000+0000/
    │       │   └── part-0.parquet
    │       └── as_of=2023-01-02T120000+0000/
    │           └── part-0.parquet
    └── data_type=NONE_AT/
        └── dataset=my_dataset/
            └── as_of=__HIVE_DEFAULT_PARTITION__/
                └── part-0.parquet
"""

from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import cast
from urllib.parse import unquote

import narwhals as nw
import pyarrow as pa
from dateutil.parser import parse
from narwhals.typing import FrameT

from .. import types
from ..config import Config
from ..dataframes import empty_frame
from ..dataframes import is_empty
from ..dataframes import merge_data
from ..dates import prepend_as_of
from ..dates import standardize_dates
from . import fs

# mypy: disable-error-code="type-var, arg-type, type-arg, return-value, attr-defined, union-attr, operator, assignment,import-untyped"

PA_TIMESTAMP_UNIT = "ns"
PA_TIMESTAMP_TZ = "UTC"
PA_FILE_FORMAT = "parquet"
PA_PARTITIONING_FLAVOR = "hive"
PA_BEHAVIOR = (
    "overwrite_or_ignore"  #  'delete_matching' did not behave as expected (?!?)
)

PA_NUMERIC = "float64"
_TIMESTAMP = pa.timestamp(unit=PA_TIMESTAMP_UNIT, tz=PA_TIMESTAMP_TZ)  # type: ignore[call-overload]


class HiveFileSystem:
    """A filesystem abstraction for reading and writing Hive-partitioned datasets."""

    def __init__(
        self,
        repository: Any,
        set_name: str,
        set_type: types.SeriesType,
        as_of_utc: datetime | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        """Initialize the filesystem handler for a given dataset."""
        if isinstance(repository, dict):
            self.repository = repository
        else:
            cfg = Config.active()
            self.repository = cfg.repositories.get(repository)

        self.set_name = set_name
        self.data_type = set_type

        if as_of_utc is None and set_type.versioning == types.Versioning.AS_OF:
            raise ValueError(
                "An 'as of' datetime must be specified when the type has versioning of type Versioning.AS_OF."
            )
        self.as_of_utc = as_of_utc

    @property
    def root(self) -> str:
        """Return the root path of the configured repository."""
        ts_root = self.repository["directory"]["options"]["path"]
        return str(ts_root)

    @property
    def directory(self) -> str:
        """Return the data directory for the dataset."""
        return str(
            Path(self.root)
            / f"data_type={self.data_type.versioning!s}_{self.data_type.temporality!s}"
            / f"dataset={self.set_name}"
        )

    def read(self, *args, **kwargs) -> FrameT:
        """Read a partitioned dataset from the filesystem."""
        if not self.exists:
            return empty_frame()

        # Define the full schema, including the partition key, to avoid type inference errors
        # when the partition only contains nulls (as is the case for Versioning.NONE).
        # partition_schema = pa.schema( [pa.field("as_of", pa.timestamp("ns", tz="UTC"), nullable=True)])

        (_file_schema, partitioning) = _parquet_schema(
            self.data_type,
            {
                "name": self.set_name,
                "versioning": self.data_type.versioning,
                "temporality": self.data_type.temporality,
            },  # define minimal tag dict, because it can not be empty (TODO: relax that?)
            partition_by=["as_of"],
        )

        dataset = pa.dataset.dataset(  # type: ignore[call-overload]
            self.directory,
            format=PA_FILE_FORMAT,
            partitioning=partitioning,
            partition_base_dir=self.directory,
        )
        table = dataset.to_table()

        # The 'as_of' column is a storage detail and should not be part of the logical dataset
        if (
            "as_of" in table.column_names
            and self.data_type.versioning == types.Versioning.NONE
        ):
            table = table.drop(["as_of"])

        return table

    def write(self, data: FrameT, tags: dict | None = None) -> None:
        """Write data to the filesystem, partitioned by versioning scheme."""
        df = prepend_as_of(data, self.as_of_utc)
        df = standardize_dates(df)
        (file_schema, partitioning) = _parquet_schema(
            self.data_type,
            tags,
            partition_by=["as_of"],
        )
        if self.data_type.versioning == types.Versioning.NONE:
            old_data = self.read()
            if not is_empty(old_data):
                old_data = prepend_as_of(old_data, None)
                df = merge_data(
                    old=old_data,
                    new=df,
                    date_cols=set(self.data_type.date_columns),
                    temporality=self.data_type.temporality,
                )

        pa_table = nw.from_native(df).to_arrow()
        pa_table = pa_table.select(file_schema.names).cast(file_schema)

        pa.dataset.write_dataset(
            pa_table,
            base_dir=self.directory,
            partitioning=partitioning,
            existing_data_behavior=PA_BEHAVIOR,
            format=PA_FILE_FORMAT,
            schema=file_schema,
        )

    @property
    def exists(self) -> bool:
        """Check if the dataset directory exists."""
        return fs.exists(self.directory)

    def versions(self) -> list[datetime | str]:
        """List available versions by inspecting subdirectories."""
        if not self.exists or self.data_type.versioning != types.Versioning.AS_OF:
            return []

        version_dirs = fs.ls(self.directory)
        versions = []
        for d in version_dirs:
            if "as_of=" in d:
                version_str = d.split("as_of=")[-1]
                versions.append(parse(unquote(version_str)))
        return sorted(versions)


def _partitioning_schema(
    file_schema: pa.Schema,
    partition_by: list[str],
    flavor: str = PA_PARTITIONING_FLAVOR,
):
    """Return a partitioning schema from pa.Table.schema."""
    part_fields = [field for field in file_schema if field.name in partition_by]
    return pa.dataset.partitioning(pa.schema(part_fields), flavor=cast(str, flavor))  # type: ignore[call-overload]


def _parquet_schema(
    data_type: types.SeriesType,
    meta: dict[str, Any],
    partition_by: list[str],
) -> tuple[pa.Schema, pa.dataset.Partitioning]:
    """Translate dataset tags into a PyArrow schema with embedded metadata."""
    from .json_helpers import tags_to_json

    if not meta:
        raise ValueError("Tags can not be empty.")

    dataset_meta = deepcopy(meta)
    series_meta = dataset_meta.pop("series", {})

    date_col_fields = [
        pa.field(
            d,
            _TIMESTAMP,  # type: ignore[call-overload]
            nullable=True,
        )
        for d in set(data_type.date_columns) | {"as_of"}
    ]
    date_col_fields.sort(key=lambda x: x.name)

    if not series_meta:
        num_col_fields = [
            pa.field(
                series_key,
                PA_NUMERIC,
                nullable=True,
                metadata=tags_to_json(series_tags),
            )
            for series_key, series_tags in series_meta.items()
        ]
        num_col_fields.sort(key=lambda x: x.name)
    else:
        num_col_fields = []

    schema = pa.schema(
        date_col_fields + num_col_fields,
        metadata=cast(dict[str, str], tags_to_json(dataset_meta)),
    )
    partition_by_fields = [f for f in schema if f.name in partition_by]
    partitioning = pa.dataset.partitioning(
        pa.schema(partition_by_fields),
        flavor=cast(str, PA_PARTITIONING_FLAVOR),  # type: ignore[call-overload]
    )
    return (schema, partitioning)



"""Use this module for running / validating setups?

ISSUE: the code is not found when trying to run `python timeseries`
... but is working fine for `python -m timeseries`
--> have a closer look at pyproject.toml?
"""

from __future__ import annotations

# from ssb_timeseries import config
import click

from ssb_timeseries.config import Config


@click.command()
def main() -> None:
    """Validate and print the active configuration."""
    active_config = Config.active()
    print(active_config)
    # perform set up steps:
    # os.environ["TIMESERIES_CONFIG"] = DEFAULT_CONFIG_LOCATION


if __name__ == "__main__":
    """Running `python timeseries` or `python -m timeseries` should run or validate setup."""
    main()



"""Provides a function for creating a PyArrow schema with embedded metadata."""

from __future__ import annotations

from copy import deepcopy
from typing import Any
from typing import cast

import pyarrow

from .. import types
from .json_helpers import tags_to_json

# mypy: disable-error-code="arg-type"

# TODO: get this from config / dataset metadata:
PA_TIMESTAMP_UNIT = "ns"
PA_TIMESTAMP_TZ = "UTC"
PA_NUMERIC = "float64"

_TIMESTAMP = pyarrow.timestamp(unit=PA_TIMESTAMP_UNIT, tz=PA_TIMESTAMP_TZ)  # type: ignore[call-overload]


def parquet_schema(
    data_type: types.SeriesType,
    meta: dict[str, Any],
) -> pyarrow.Schema | None:
    """Translate dataset tags into a PyArrow schema with embedded metadata."""
    if not meta:
        raise ValueError("Tags can not be empty.")

    dataset_meta = deepcopy(meta)
    series_meta = dataset_meta.pop("series")

    if not series_meta:
        return None

    date_col_fields = [
        pyarrow.field(
            d,
            _TIMESTAMP,  # type: ignore[call-overload]
            nullable=False,
        )
        for d in set(data_type.date_columns)
    ]

    num_col_fields = [
        pyarrow.field(
            series_key,
            PA_NUMERIC,
            nullable=True,
            metadata=tags_to_json(series_tags),
        )
        for series_key, series_tags in series_meta.items()
    ]
    date_col_fields.sort(key=lambda x: x.name)
    num_col_fields.sort(key=lambda x: x.name)

    schema = pyarrow.schema(
        date_col_fields + num_col_fields,
        metadata=cast(dict[str, str], tags_to_json(dataset_meta)),
    )
    return schema



"""Intervals are crucial to data retrieval.

The most used are likely to be the trivial 'all' and 'last',
but more sophisticated ranges may be specified using all date attributes for a given series type.
Advanced cases may even blur the line to date arithmetics.
An extended interval specification could specify filters, eg 'if day()==Monday',
or time steps or windows relative to another date atribute, eg 'valid_at == as_of + 2 months'.
"""

from __future__ import annotations

from datetime import datetime

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self  # noqa: UP035 #backport to 3.10

import ssb_timeseries as ts

# mypy: disable-error-code="no-untyped-def,return-value,no-any-return,attr-defined,operator"


class Interval:
    """Intervals specify date ranges by way of inclusive 'start' and exclusive 'stop' datetimes."""

    def __init__(
        self, f: datetime | None = None, t: datetime | None = None, **kwargs
    ) -> None:
        """Interval(date_from, date_to).

        A number of variations of named parameters: start/stop, begin/end, as_of, as_of_from/as_of_to, valid_from, valid_to -
        If only "as_of" is provided, start/stop are both set to this date.
        Interval.start defaults to datetime.min
        Interval.stop defaults to datetime.max
        """
        as_of = kwargs.get("as_of")
        as_of_from = kwargs.get("as_of_from")
        as_of_to = kwargs.get("as_of_to")
        valid_from = kwargs.get("valid_from")
        valid_to = kwargs.get("valid_to")
        start = kwargs.get("start")
        stop = kwargs.get("stop")
        from_ = kwargs.get("from_date")
        to = kwargs.get("to_date")
        date_from = kwargs.get("date_from")
        date_to = kwargs.get("date_to")
        from_date = kwargs.get("from_date")
        to_date = kwargs.get("to_date")
        begin = kwargs.get("begin")
        end = kwargs.get("end")

        precision = kwargs.get("precision", ts.dates.MAX_TIME_PRECISION)

        ts.logger.debug(f"Interval.__init__ with kwargs:\n{kwargs}")

        self.start = (
            f
            or as_of
            or start
            or as_of_from
            or from_
            or date_from
            or from_date
            or valid_from
            or begin
            or datetime.min
        )
        self.stop = (
            t
            or as_of
            or stop
            or to
            or as_of_to
            or date_to
            or to_date
            or valid_to
            or end
            or datetime.max
        )

        if precision:
            # "round" to MAX_TIME_precision if provided
            ...

    def includes(self, *args: datetime | list[datetime]) -> bool | list[bool]:
        """Check if one or more dates are included in interval."""
        ts.logger.debug(f"Interval.include args: {args}")
        if len(args) == 1:
            ts.logger.debug("Interval.include - single input")
            out = args[0] >= self.start and args[0] <= self.stop
        else:
            out = [x >= self.start and x <= self.stop for x in args]
        ts.logger.debug(f"Interval.include returns:\n{out}")
        return out

    def all(self) -> None:
        """Set start and stop to widest possible span."""
        self.start = datetime.min
        self.stop = datetime.max

    def __eq__(self, other: object) -> bool:
        """Check if two intervals are equal."""
        return (self.start == other.start) and (self.stop == other.stop)

    def __lt__(self, other: Self) -> bool:
        """Check if the entire left side interval is before and outside the right side interval."""
        ts.logger.debug(
            f"Interval.lt: \n\t{self.stop} < ({other.start} < {other.stop})"
        )
        return self.stop < other.start

    def __gt__(self, other: Self) -> bool:
        """Check if the entire left side interval is after and outside the right side interval."""
        ts.logger.debug(f"Interval.gt: \n\t{self.start} > {other.stop}")
        return self.start > other.stop

    def __le__(self, other: Self) -> bool:
        """Check if the left side interval begins before and ends before or inside the right side interval."""
        ts.logger.debug(
            f"Interval.le: \n\t{self.start} - {self.stop}\n\t{other.start} - {other.stop}"
        )
        return (self.start, self.stop) <= (other.start, other.stop)

    def __ge__(self, other: Self) -> bool:
        """Check if the left side interval begins after or inside and ends after the right side interval."""
        ts.logger.debug(
            f"Interval.ge: \n\t{self.start} - {self.stop}\n\t{other.start} - {other.stop}"
        )
        return (self.start, self.stop) >= (other.start, other.stop)

    def __repr__(self) -> str:
        """Return a string representation able to reinitialize the interval."""
        expr = f"Interval(from_date = {self.start}, to_date = {self.stop}) "
        return expr

    def __str__(self) -> str:
        """Return a string representation of the interval."""
        expr = f"Interval from {self.start} to {self.stop}."
        return expr



from __future__ import annotations

import random
import re
from typing import Any

from ssb_timeseries.meta import Taxonomy

_words: list[str] = [
    "apple",
    "book",
    "desk",
    "pen",
    "cat",
    "dog",
    "tree",
    "house",
    "car",
    "phone",
    "computer",
    "laptop",
    "keyboard",
    "mouse",
    "chair",
    "table",
    "door",
    "window",
    "wall",
    "floor",
]


_aggregates: list[dict[str, Any]] = [
    {"code": "count", "name": "count", "parent": None},
    {"code": "sum", "name": "sum", "parent": None},
    {"code": "avg", "name": "avg", "parent": None},
    {"code": "min", "name": "min", "parent": None},
    {"code": "max", "name": "min", "parent": None},
    {"code": "perc01", "name": "perc01", "parent": None},
    {"code": "perc05", "name": "perc05", "parent": None},
    {"code": "perc10", "name": "perc10", "parent": None},
    {"code": "perc15", "name": "perc15", "parent": None},
    {"code": "perc20", "name": "perc20", "parent": None},
    {"code": "perc25", "name": "perc25", "parent": None},
    {"code": "perc30", "name": "perc30", "parent": None},
    {"code": "perc35", "name": "perc35", "parent": None},
    {"code": "perc40", "name": "perc40", "parent": None},
    {"code": "perc45", "name": "perc45", "parent": None},
    {"code": "perc50", "name": "perc50", "parent": None},
    {"code": "perc55", "name": "perc55", "parent": None},
    {"code": "perc60", "name": "perc60", "parent": None},
    {"code": "perc65", "name": "perc65", "parent": None},
    {"code": "perc70", "name": "perc70", "parent": None},
    {"code": "perc75", "name": "perc75", "parent": None},
    {"code": "perc80", "name": "perc80", "parent": None},
    {"code": "perc85", "name": "perc85", "parent": None},
    {"code": "perc90", "name": "perc90", "parent": None},
    {"code": "perc95", "name": "perc95", "parent": None},
    {"code": "perc99", "name": "perc99", "parent": None},
]

_nordic: list[dict[str, Any]] = [
    {"code": "nord", "name": "Nordic countries", "parent": None},
    {"code": "scan", "name": "Nordic countries", "parent": "nord"},
    {"code": "DK", "name": "Denmark", "parent": "scan"},
    {"code": "SE", "name": "Consumption", "parent": "scan"},
    {"code": "NO", "name": "Export", "parent": "scan"},
    {"code": "FI", "name": "Injections", "parent": "nord"},
    {"code": "IS", "name": "Supply", "parent": "nord"},
]
# Indclude province level?
# Sweden:
# SE-K	Blekinge län	Blekinge	[SE-10]
# SE-W	Dalarnas län	Dalarna	[SE-20]
# SE-I	Gotlands län	Gotland	[SE-09]
# SE-X	Gävleborgs län	Gävleborg	[SE-21]
# SE-N	Hallands län	Halland	[SE-13]
# SE-Z	Jämtlands län	Jämtland	[SE-23]
# SE-F	Jönköpings län	Jönköping	[SE-06]
# SE-H	Kalmar län	Kalmar	[SE-08]
# SE-G	Kronobergs län	Kronoberg	[SE-07]
# SE-BD	Norrbottens län	Norrbotten	[SE-25]
# SE-M	Skåne län	Scania	[SE-12]
# SE-AB	Stockholms län	Stockholm	[SE-01]
# SE-D	Södermanlands län	Södermanland	[SE-04]
# SE-C	Uppsala län	Uppsala	[SE-03]
# SE-S	Värmlands län	Värmland	[SE-17]
# SE-AC	Västerbottens län	Västerbotten	[SE-24]
# SE-Y	Västernorrlands län	Western Northland	[SE-22]
# SE-U	Västmanlands län	Västmanland	[SE-19]
# SE-O	Västra Götalands län	Västra Götaland	[SE-14]
# SE-T	Örebro län	Örebro	[SE-18]
# SE-E	Östergötlands län	Östergötland	[SE-05]

_econ: list[dict[str, str | None]] = [
    {"code": "B", "name": "Balance", "parent": None},
    {"code": "D", "name": "Demand", "parent": "B"},
    {"code": "C", "name": "Consumption", "parent": "D"},
    {"code": "Exp", "name": "Export", "parent": "D"},
    {"code": "Inj", "name": "Injections", "parent": "D"},
    {"code": "S", "name": "Supply", "parent": "B"},
    {"code": "P", "name": "Production", "parent": "S"},
    {"code": "Imp", "name": "Imports", "parent": "S"},
    {"code": "Wdr", "name": "Withdrawals", "parent": "S"},
]
"""Sample economic variables."""

_commodity: list[dict[str, str | None]] = [
    {"code": "petro", "name": "petroleum products", "parent": None},
    {"code": "oil", "name": "oil", "parent": "petro"},
    {"code": "gas", "name": "natural gas", "parent": "petro"},
    {"code": "lpg", "name": "liquefied petroleum gas", "parent": "petro"},
]
"""A list of petroleum products."""

_bio: list[dict[str, str | None]] = [
    {"code": "animals", "name": "animals", "parent": None},
    {"code": "plants", "name": "plants", "parent": None},
    {"code": "fungi", "name": "fungi", "parent": None},
    {"code": "microbes", "name": "microbes", "parent": None},
]
"""A list of biological categories."""


def random_words(some_list: list[str] | None = None, n: int = 1) -> list[str]:
    """Return a random word from 'some_list'."""
    if not some_list:
        some_list = _words
    return random.choices(some_list, k=n)


def tags_string(separator: str = "_", *args: list[str]) -> str:
    """Join one random element from each list in '*args*', separated by 'separator'."""
    return separator.join([random.choice(a) for a in args])


def lower_first_char(s: str) -> str:
    """Change first character of string to lowercase."""
    return s[0].lower() + s[1:]


def camel_case_keys(dictionary: dict) -> dict:
    """Rename dictionary keys to camel case."""
    out = {}

    for k, v in dictionary:
        pascal_k = "".join(word.title() for word in k.split("_"))
        camel_k = lower_first_char(pascal_k)
        if isinstance(v, dict):
            out[camel_k] = camel_case_keys(v)
        else:
            out[camel_k] = v
    return out


def snake_case_keys(dictionary: dict) -> dict:
    """Rename dictionary keys to snake case."""
    out = {}
    for k, v in dictionary:
        snake_k = re.sub(r"(?<!^)(?=[A-Z])", "_", k).lower()
        if isinstance(v, dict):
            out[snake_k] = snake_case_keys(v)
        else:
            out[snake_k] = v
    return out


def list_to_tax(list_of_dicts: list[dict[str, Any]]) -> list[dict[str, Any]]:
    """Use code as key, convert list of dicts to dict."""
    ...
    blanks = ["shortName", "presentationName", "validFrom", "validTo", "notes"]
    # dict_of_dicts = {d['code']: d for d in list_of_dicts}
    # for k in dict_of_dicts.keys():
    for d in list_of_dicts:
        # dict_of_dicts[k]['parentCode'] =  dict_of_dicts[k].pop('parent','')
        d["parentCode"] = d.pop("parent", "")
        for b in blanks:
            d[b] = ""
    return list_of_dicts


def balance() -> Taxonomy:
    """Return sample economic balance taxonomy."""
    d = list_to_tax(_econ)
    return Taxonomy(data=d)


def commodity() -> Taxonomy:
    """Return sample economic balance taxonomy."""
    d = list_to_tax(_commodity)
    return Taxonomy(data=d)


def nordic_countries() -> Taxonomy:
    """Return sample economic balance taxonomy."""
    d = list_to_tax(_nordic)
    return Taxonomy(data=d)


def aggregates() -> Taxonomy:
    """Return sample economic balance taxonomy."""
    d = list_to_tax(_aggregates)
    return Taxonomy(data=d)


# def words() -> Taxonomy:
#     """Return sample economic balance taxonomy."""
#     return Taxonomy(data=_words)


def biology() -> Taxonomy:
    """Return sample economic balance taxonomy."""
    d = list_to_tax(_bio)
    return Taxonomy(data=d)



"""Provides a PyArrow-based simple, file-based I/O handler for Parquet format.

This handler stores datasets in a wide format (series as columns) with embedded metadata,
using a defined directory structure. For example:

.. code-block::

    <repository_root>/
    ├── AS_OF_AT/
    │   └── my_versioned_dataset/
    │       ├── my_versioned_dataset-as_of_20230101T120000+0000-data.parquet
    │       └── my_versioned_dataset-as_of_20230102T120000+0000-data.parquet
    └── NONE_AT/
        └── my_dataset/
            └── my_dataset-latest-data.parquet

It uses PyArrow for eager reading and writing.
"""

from __future__ import annotations

import os
import re
from datetime import datetime
from typing import Any
from typing import NamedTuple
from typing import cast

import narwhals as nw
import pyarrow
import pyarrow.compute
from narwhals.typing import FrameT

from .. import types
from ..config import Config
from ..dataframes import empty_frame
from ..dataframes import is_empty
from ..dataframes import merge_data
from ..dates import date_utc
from ..dates import datelike_to_utc
from ..dates import prepend_as_of
from ..dates import utc_iso_no_colon
from ..logging import logger
from ..types import PathStr
from . import fs
from .parquet_schema import parquet_schema

# mypy: disable-error-code="type-var, arg-type, type-arg, return-value, attr-defined, union-attr, operator, assignment,import-untyped, "


active_config = Config.active

# TODO: get this from config / dataset metadata:
PA_TIMESTAMP_UNIT = "ns"
PA_TIMESTAMP_TZ = "UTC"
PA_NUMERIC = "float64"


def _version_from_file_name(
    file_name: str, pattern: str | types.Versioning = "as_of", group: int = 2
) -> str:
    """Extract a version marker from a filename using known patterns."""
    if isinstance(pattern, types.Versioning):
        pattern = str(pattern)

    match pattern.lower():
        case "persisted":
            regex = r"(_v)(\d+)(.parquet)"
        case "as_of":
            date_part = "[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{6}[+-][0-9]{4}"
            regex = f"(as_of_)({date_part})(-data.parquet)"
        case "names":
            # type is not implemented
            regex = "(_v)(*)(-data.parquet)"
        case "none":
            regex = "(.*)(latest)(-data.parquet)"
        case _:
            regex = pattern

    vs = re.search(regex, file_name).group(group)
    logger.debug(
        "file: %s pattern:%s, regex%s \n--> version: %s ",
        file_name,
        pattern,
        regex,
        vs,
    )
    return vs


def last_version_number_by_regex(directory: str, pattern: str = "*") -> str:
    """Return the max version number from files in a directory matching a pattern."""
    files = fs.ls(directory, pattern=pattern)
    number_of_files = len(files)

    vs = sorted([int(_version_from_file_name(fname, "persisted")) for fname in files])
    if vs:
        read_from_filenames = max(vs)
        out = read_from_filenames
    else:
        read_from_filenames = 0
        out = number_of_files

    logger.debug(
        "io/pyarrow_simple.last_version_number_by_regex() search in directory: \n\t%s\n\tfor '%s' found %s files, regex identified version %s --> vs %s.",
        directory,
        pattern,
        f"{number_of_files!s}",
        f"{read_from_filenames!s}",
        f"{out!s}",
    )
    return out


class FileSystem:
    """A filesystem abstraction for reading and writing dataset data."""

    def __init__(
        self,
        repository: Any,  # dict[str,str] | FileBasedRepository,
        set_name: str,
        set_type: types.SeriesType,
        as_of_utc: datetime | None = None,
        process_stage: str = "statistikk",
        sharing: dict | None = None,
    ) -> None:
        """Initialize the filesystem handler for a given dataset.

        This method calculates the necessary directory structure based on the
        dataset's type and name.
        """
        if isinstance(repository, dict):
            self.repository = repository
        else:
            cfg = Config.active()
            self.repository = cfg.repositories.get(repository)

        self.set_name = set_name
        self.data_type = set_type

        self.process_stage = process_stage
        self.sharing = sharing

        if as_of_utc is None and set_type.versioning == types.Versioning.AS_OF:
            raise ValueError(
                "An 'as of' datetime must be specified when the type has versioning of type Versioning.AS_OF."
            )

        self.as_of_utc: datetime = as_of_utc

    @property
    def root(self) -> str:
        """Return the root path of the configured repository."""
        ts_root = self.repository["directory"]["options"]["path"]
        return str(ts_root)

    @property
    def filename(self) -> str:
        """Construct the standard filename for the dataset's data file."""
        match str(self.data_type.versioning):
            case "AS_OF":
                safe_timestamp = utc_iso_no_colon(self.as_of_utc)
                file_name = f"{self.set_name}-as_of_{safe_timestamp}-data.parquet"
            case "NONE":
                file_name = f"{self.set_name}-latest-data.parquet"
            case "NAMED":
                file_name = f"{self.set_name}-NAMED-data.parquet"
            case _:
                raise ValueError("Unhandled versioning.")

        logger.debug(file_name)
        return file_name

    @property
    def directory(self) -> str:
        """Return the data directory for the dataset."""
        return os.path.join(
            self.root,
            f"{self.data_type.versioning!s}_{self.data_type.temporality!s}",
            self.set_name,
        )

    @property
    def fullpath(self) -> str:
        """Return the full path to the dataset's data file."""
        return os.path.join(self.directory, self.filename)

    def read(
        self,
        interval: str = "",  # TODO: Implement use av interval = Interval.all,
    ) -> pyarrow.Table:
        """Read data from the filesystem.

        Returns an empty dataframe if the file is not found.
        """
        logger.debug(interval)
        if fs.exists(self.fullpath):
            logger.info(
                "DATASET.read.start %s: Reading data from file %s",
                self.set_name,
                self.fullpath,
            )
            try:
                df = fs.read_parquet(self.fullpath, implementation="pyarrow")
                logger.info("DATASET.read.success %s: Read data.", self.set_name)
            except FileNotFoundError:
                logger.exception(
                    "DATASET.read.error %s: Read data failed. File not found: %s",
                    self.set_name,
                    self.fullpath,
                )
                df = empty_frame()

        else:
            df = empty_frame()
            logger.debug(f"No file {self.fullpath} - return empty frame instead.")
        pa_table = datelike_to_utc(df)

        # The 'as_of' column is a storage detail and should not be part of the logical dataset
        if "as_of" in pa_table.column_names:
            pa_table = pa_table.drop(["as_of"])

        return cast(pyarrow.Table, pa_table)

    def write(self, data: FrameT, tags: dict | None = None) -> None:
        """Write data to the filesystem.

        If versioning is AS_OF, a new file is always created.
        If versioning is NONE, new data is merged into the existing file.
        """
        new = nw.from_native(data)
        if self.data_type.versioning == types.Versioning.AS_OF:
            # consider a merge option for versioned writing?
            df = prepend_as_of(new, self.as_of_utc)
        else:
            old = self.read(self.set_name)
            if is_empty(old):
                df = new
            else:
                logger.debug(
                    f"Merging data with temporality: {self.data_type.temporality}"
                )
                logger.debug(f"Old data length: {len(old)}")
                logger.debug(f"New data length: {len(new)}")
                df = merge_data(
                    new=new,
                    old=old,
                    date_cols=self.data_type.date_columns,
                    temporality=self.data_type.temporality,
                )
                logger.debug(f"Merged data length: {len(df)}")

        logger.info(
            "DATASET.write.start %s: writing data to file\n\t%s\nstarted.",
            self.set_name,
            self.fullpath,
        )
        try:
            fs.write_parquet(
                data=df,
                path=self.fullpath,
                schema=parquet_schema(self.data_type, tags),
            )
        except Exception as e:
            logger.exception(
                "DATASET.write.error %s: writing data to file\n\t%s\nreturned exception: %s.",
                self.set_name,
                self.fullpath,
                e,
            )
        logger.info(
            "DATASET.write.success %s: writing data to file\n\t%s\nended.",
            self.set_name,
            self.fullpath,
        )

    @property
    def exists(self) -> bool:
        """Check if the data file for the dataset exists."""
        return fs.exists(self.fullpath)

    def versions(
        self, file_pattern: str = "*", pattern: str | types.Versioning = "as_of"
    ) -> list[datetime | str]:
        """List all available version markers from the data directory."""
        files = fs.ls(self.directory, pattern=file_pattern)
        versions: list[str | datetime] = []
        if files:
            vs_strings = [
                _version_from_file_name(str(fname), pattern, group=2) for fname in files
            ]
            match types.Versioning(pattern):
                case types.Versioning.AS_OF:
                    versions = sorted([date_utc(as_of) for as_of in vs_strings])
                case types.Versioning.NAMES:
                    versions = sorted(vs_strings)
                case types.Versioning.NONE:
                    versions = vs_strings
                case _:
                    raise ValueError(f"pattern '{pattern}' not recognized.")
        return versions


# ================================ SEARCH: =================================


class SearchResult(NamedTuple):
    """Represents a single item in a search result."""

    name: str
    type_directory: str


def find_datasets(
    pattern: str | PathStr = "",
    exclude: str = "metadata",
    repository: list[PathStr] | PathStr = "",
) -> list[SearchResult]:
    """Search for dataset directories in all configured repositories.

    Args:
        pattern: A glob pattern to match against directory names.
        exclude: A substring to exclude from the search results.
        repository: A specific repository path to search in. If empty,
            searches all configured repositories.

    Returns:
        A list of SearchResult objects for the found datasets.
    """
    if pattern:
        pattern = f"*{pattern}*"
    else:
        pattern = "*"

    if repository:
        search_directories = [repository]
        repo_names = ["root"]
        logger.debug("IO.find_dataset pattern %s in repo %s", pattern, repository)
    else:
        search_directories = [
            v["directory"]["options"]["path"]
            for k, v in active_config().repositories.items()
        ]
        repo_names = list(active_config().repositories.keys())

    directories = []
    for search_dir in search_directories:
        directories.extend(fs.find(search_dir, pattern, full_path=True))

    logger.debug("%s %s", pattern, directories)
    if exclude:
        dirs = [d for d in directories if exclude not in d]
        logger.debug(
            "DATASET.IO.find_datasets: exclude '%s' eliminated:\n%s",
            exclude,
            [d for d in dirs if exclude in d],
        )
    search_results = []
    for search_dir, repo in zip(search_directories, repo_names, strict=False):
        logger.debug("%s | %s", search_dir, repo)
        search_results.extend(
            [d.replace(search_dir, repo).split(os.path.sep) for d in dirs]
        )
    logger.debug("search results: %s", search_results)
    return [SearchResult(f[2], f[1]) for f in search_results]



"""Configurations for the SSB timeseries library.

An environment variable TIMESERIES_CONFIG is expected to point to a JSON file with configurations.
If these exist, they will be loaded and put into a Config object CONFIG when the configuration module is loaded.

In most cases, this would happen behind the scene when :py:mod:`ssb_timeseries.dataset` or :py:mod:`ssb_timeseries.catalog` are imported.

Directly accessing the configuration module should only be required when manipulating configurations from Python code.

Example:
    >>> # doctest: +SKIP
    >>> from ssb_timeseries.config import CONFIG
    >>> CONFIG.catalog = 'gs://{bucket}/timeseries/metadata/'
    >>> CONFIG.save()
    >>> # doctest: -SKIP

For switching between preset configurations, use the `timeseries-config` command::

    poetry run timeseries-config <option>

which is equivalent to::

    python ./config.py <option>

See :py:func:`ssb_timeseries.config.main` for details on the named options.
"""

from __future__ import annotations

import json
import logging
import os
import sys
import warnings
from pathlib import Path

try:
    from typing import NotRequired
    from typing import Required
    from typing import Self
    from typing import TypedDict
except ImportError:
    from typing_extensions import NotRequired  # noqa: UP035 #backport to 3.10
    from typing_extensions import Required  # noqa: UP035 #backport to 3.10
    from typing_extensions import Self  # noqa: UP035 #backport to 3.10
    from typing_extensions import TypedDict

from typing import Any
from typing import TypeAlias

from .types import PathStr

# mypy: disable-error-code="assignment, arg-type, override,call-arg,has-type,no-untyped-def,attr-defined,import-untyped,"

_config_logger = logging.getLogger(__name__)
PACKAGE_NAME = "ssb_timeseries"
ENV_VAR_NAME = "TIMESERIES_CONFIG"


class FileRepoConfig(TypedDict):
    """Links a path and a IO handler for a single file based repository."""

    handler: Required[str]
    options: Required[dict[str, Any]]


class Repository(TypedDict):
    """Defines data and metadata handling for time series repositories."""

    name: NotRequired[str]
    directory: Required[FileRepoConfig]
    catalog: NotRequired[FileRepoConfig]
    default: NotRequired[bool]


FileBasedRepository: TypeAlias = Repository


class ConfigDict(TypedDict):
    """Required attributes for configuration."""

    configuration_file: Required[str]
    io_handlers: Required[dict[str, Any]]
    repositories: Required[dict[str, Repository]]
    snapshots: NotRequired[dict[str, Repository]]
    sharing: NotRequired[dict[str, Repository]]
    log_file: NotRequired[str]
    logging: Required[dict[str, Any]]


def is_valid_config(configuration: ConfigDict) -> tuple[bool, object]:
    """Check if a dictionary is a valid configuration :py:class:`ConfigDict`."""
    # The ConfigDict.__required_keys__ includes optional fields like 'snapshots' and 'sharing'
    # which causes a ValidationError when the default configuration is loaded.
    # To fix this, we explicitly define the required keys.
    # missing_required = ConfigDict.__required_keys__ - set(configuration.keys())
    required_keys = {"configuration_file", "io_handlers", "repositories", "logging"}
    missing_required = required_keys - set(configuration.keys())
    if missing_required:
        msg = f"Configuration is missing required fields: {list(missing_required)}\n{configuration}"
        return (False, msg)

    wrong_type = []
    for (
        cfg_key,
        cfg_expected_type,
    ) in ConfigDict().items():  # type: ignore [typeddict-item]
        config_item = configuration.get(cfg_key, None)
        cfg_got_type = type(config_item)
        if cfg_got_type is type(cfg_expected_type):
            wrong_type.append(
                f"{cfg_key} - got {cfg_got_type} - expected {cfg_expected_type}"
            )

    if wrong_type:
        msg = f"Configuration fields have wrong type: {wrong_type}"
        return (False, msg)

    return (True, None)


def unset_env_var() -> str:
    """Unsets the environment variable :py:const:`ENV_VAR_NAME` and returns the value that was unset."""
    return os.environ.pop(ENV_VAR_NAME, "")


def active_file(path: PathStr = "") -> str:
    """If a path is provided, sets environment variable :py:const:`ENV_VAR_NAME` to specify the location of the configuration file.

    Returns the value of the environment variable.
    """
    if path:
        os.environ[ENV_VAR_NAME] = str(path)
        _config_logger.debug(f"Set environment variable {ENV_VAR_NAME} to {path}")

    return os.environ.get(ENV_VAR_NAME, "")


HOME = str(Path.home())

DAPLALAB_WORK = "/home/onyxia/work"
DAPLALAB_FUSE = "/buckets"
SSB_DIR_NAME = "tidsserier"
ROOT_DIR_NAME = "timeseries"
META_DIR_NAME = "metadata"
SSB_CONF_DIR = "konfigurasjon"
LINUX_CONF_DIR = ".config"
SSB_LOGDIR = "logger"
LOGDIR = "logs"
LOGFILE = "timeseries.log"
CONFIGFILE = "timeseries_config.json"

DAPLA_TEAM_CONTEXT = os.getenv("DAPLA_TEAM_CONTEXT", "")
DAPLA_ENV = os.getenv("DAPLA_ENVIRONMENT", "")
"""Returns the Dapla environment: 'prod' | test | dev"""
DAPLA_TEAM = os.getenv("DAPLA_TEAM", "<teamname>")
"""Returns the Dapla team/project name.'"""
DAPLA_BUCKET = f"gs://{DAPLA_TEAM}-{DAPLA_ENV}"
"""Returns the Dapla product bucket name for the current environment: gs://{DAPLA_TEAM}-{DAPLA_ENV}."""

LOGGING_PRESETS = {
    "simple": {
        "version": 1,
        "disable_existing_loggers": False,
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
                "level": "INFO",
            },
        },
        "loggers": {
            PACKAGE_NAME: {
                "level": "INFO",
                "handlers": ["console"],
                "propagate": False,
            }
        },
    },
    "console+file": {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "format": "%(asctime)s - %(levelname)s - %(message)s",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
            "json": {
                "format": '{"time": %(asctime)-s, "level": %(levelname)-s, "message": %(message)s},',
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
        },
        "handlers": {
            "console": {
                "level": "INFO",
                "class": "logging.StreamHandler",
                "formatter": "default",
                "stream": "ext://sys.stdout",
            },
            "file": {
                "level": "INFO",
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "default",
                "filename": "ssb_timeseries.log",
                "maxBytes": 10_000,
                "backupCount": 3,
            },
        },
        "loggers": {
            PACKAGE_NAME: {
                "level": "INFO",
                "handlers": ["console", "file"],
                "propagate": False,
            }
        },
    },
}

BUILTIN_IO_HANDLERS = {
    "simple-parquet": {
        "handler": "ssb_timeseries.io.pyarrow_simple.FileSystem",
        "options": {},
    },
    "json": {
        "handler": "ssb_timeseries.io.json_metadata.JsonMetaIO",
        "options": {},
    },
    "snapshots": {
        "handler": "ssb_timeseries.io.snapshots.FileSystem",
        "options": {},
    },
}
PRESETS: dict[str, ConfigDict] = {
    "home": {
        "configuration_file": str(Path(HOME, LINUX_CONF_DIR, PACKAGE_NAME, CONFIGFILE)),
        "io_handlers": BUILTIN_IO_HANDLERS,
        "repositories": {
            DAPLA_TEAM: {
                # "name": "home",
                "directory": {
                    "handler": "parquet",
                    "options": {"path": str(Path(HOME, ROOT_DIR_NAME))},
                },
                "catalog": {
                    "handler": "json",
                    "options": {
                        "path": str(Path(HOME, ROOT_DIR_NAME, META_DIR_NAME)),
                    },
                },
            }
        },
        "log_file": str(Path(HOME, ROOT_DIR_NAME, LOGDIR, LOGFILE)),
        "logging": LOGGING_PRESETS["simple"],
    },
    "daplalab": {
        "configuration_file": str(
            Path(DAPLA_BUCKET, SSB_CONF_DIR, PACKAGE_NAME, CONFIGFILE)
        ),
        "io_handlers": BUILTIN_IO_HANDLERS,
        "repositories": {
            DAPLA_TEAM: {
                "name": DAPLA_TEAM,
                "directory": {
                    "handler": "parquet",
                    "options": {"path": str(Path(DAPLALAB_FUSE, ROOT_DIR_NAME))},
                },
                "catalog": {
                    "handler": "json",
                    "options": {
                        "path": str(Path(DAPLALAB_FUSE, SSB_DIR_NAME, META_DIR_NAME)),
                    },
                },
            }
        },
        "log_file": str(Path(DAPLALAB_FUSE, SSB_LOGDIR, LOGFILE)),
        "logging": LOGGING_PRESETS["simple"],
    },
}

PRESETS["default"] = PRESETS["home"]
PRESETS["defaults"] = PRESETS["home"]
DEFAULTS = PRESETS["default"]


class Config:
    """Configuration class; for reading and writing timeseries configurations.

    If instantiated with no parameters, an existing configuration file is exepected to exist: either in a location specified by the environment variable TIMESERIES_CONFIG or in the default location in the user's home directory. If not, an error is returned.

    If the :py:attr:`configuration_file` attribute is specified, configurations will be loaded from that file. No other parameters are required. A :py:exc:`FileNotFoundError` or :py:exc:`FileDoesNotExist` error will be returned if the file is not found. In this case, no attempt is made to load configurations from locations specified by environment variable or defaults.

    If any additional parameters are provided, they will override values from the configuration file. If the result is not a valid configuration, a ValidationError is raised.

    If one or more parameters are provided, but the `configuration_file` parameter is not among them, configurations are identified by the environment variable TIMESERIES_CONFIG or the default configuration file location (in that order of priority). Provided parameters override values from the configuration file. If the result is not a valid configuration, an error is raised.

    The returned configuration will not be saved, but held in memory only till the :py:meth:`save` method is called. Then the configuration will be savedto a file and the environment variable TIMESERIES_CONFIG set to reflect the location of the file.

    """

    configuration_file: PathStr
    """The path to the configuRation file."""
    repositories: dict[str, Repository]
    """Defines storage locations for time series data and metadata."""
    snapshots: dict[str, Repository]
    """Defines the storage locations for persisting (archiving) data in stable states."""
    sharing: dict[str, Repository]
    """Defines the storage locations for shared data."""
    io_handlers: dict[str, Any]
    """IO handlers for repository, snapshotts and sharing."""
    logging: dict[str, Any]
    """Logging configuration as a valid :py:mod:`logging.dictConfig`."""

    def __init__(self, **kwargs) -> None:  # noqa: D417, ANN003, RUF100
        """Initialize Config object from keyword arguments.

        Keyword Arguments:
            preset (str): Optional. Name of a preset configuration. If provided, the preset configuration is loaded, and no other parameters are considered.
            configuration_file (str): Path to the configuration file. If the parameter is not provided, the environment variable TIMESERIES_CONFIG is used. If the environment variable is not set, the default configuration file location is used.
            repositories (list[FileBasedRepository]): New in version 0.5.0. Replaces bucket, timeseries_root and catalog.
            log_file (str): Path to the log file.
            bucket (str): Name of the GCS bucket.
            ignore_file (bool):

        Raises:
            :py:exc:`FileNotFoundError`: If the configuration file as implied by provided or not provided parameters does not exist.   # noqa: DAR402
            :py:exc:`ValidationError`: If the resulting configuration is not valid.   # noqa: DAR402
            :py:exc:`EnvVarNotDefinedeError`: If the environment variable TIMESERIES_CONFIG is not defined.

        Examples:
            To load an existing preset configuration:

                >>> from ssb_timeseries.config import Config
                >>> config = Config(preset='daplalab')
        """
        preset_name = kwargs.pop("preset", "")
        ignore_file = kwargs.pop("ignore_file", False)
        param_specified_config_file = kwargs.get("configuration_file", "")

        kwargs_are_complete_config = is_valid_config(kwargs)[0]

        if preset_name:
            _config_logger.debug(f"Loading preset configuration {preset_name}.")
            self.apply(PRESETS[preset_name])
            return
        elif kwargs_are_complete_config:
            _config_logger.debug("Complete configuration in parameters.\n%s", kwargs)
            self.apply(kwargs)
            return
        elif param_specified_config_file:
            # if config file is
            _config_logger.info(
                f"Loading configuration from {param_specified_config_file}"
            )

            if set(kwargs.keys()) == {"configuration_file"}:
                # if config file is the only parameter, it is an error for it not to exist
                no_file_is_an_error = True
            else:
                # if kwargs form a complete config, it is ok if the config file does not exist
                # (otherwise it is needed to supplement the kwargs)
                no_file_is_an_error = not is_valid_config(kwargs)[0]

            if not ignore_file:
                config_from_file = load_json_file(
                    path=param_specified_config_file,
                    error_on_missing=no_file_is_an_error,
                )
            else:
                config_from_file = {}

            config_values = PRESETS["default"]
            config_values.update(config_from_file)  # type: ignore [typeddict-item]
            _config_logger.debug(f"FROM FILE: {config_values=}")
        elif active_file():
            # if the path is specified by the environment variable, not finding it is an error
            _config_logger.debug(f"Loading configuration from {active_file()}")
            config_values = load_json_file(
                path=active_file(),
                error_on_missing=True,
            )
        # elif not active_file():
        #    raise MissingEnvironmentVariableError
        else:
            _config_logger.warning(
                f"The environment variable {ENV_VAR_NAME} did not exist and no configuration file parameter was provided. Loading default configuration."
            )
            config_values = PRESETS["defaults"]

        config_values.update(kwargs)  # type: ignore [typeddict-item]
        self.apply(config_values)

    def apply(self, configuration: dict) -> None:
        """Set configuration values from a dictionary."""
        _config_logger.debug(f"APPLIES: {configuration=}")
        log_config = configuration.get("logging", {})
        if not log_config:
            configuration["logging"] = {}
        elif isinstance(log_config, str):
            warnings.warn(
                "string based log config! TO DO: look up named option", stacklevel=2
            )
            configuration["logging"] = {"str": log_config}

        config_ok, reason = is_valid_config(configuration=configuration)

        if not config_ok:
            _config_logger.error(f"Invalid configuration {configuration}\n{reason}.")
            raise ValidationError(f"Invalid configuration:\n{configuration}\n{reason}.")

        logfile = configuration.pop("log_file", "")
        if logfile and not logging:
            # TODO: filehandler should be configured as dictConfig
            # .. and we should not enter this block?
            # --> TODO: Check / remove OR add dictConfig for the following:
            configuration["logging"] = {"logfile": logfile}
        else:
            ...
            # --- if logging is valid logging.dictConfig -->
            # (add file handler first?)
            # handlers = configuration['logging'].get('handlers',{})
            # filehandler =  handlers.get('file',{})
            # if logfile and filehandler:
            #     configuration['logging']['handlers']['file']['filename'] = logfile

        for key, value in configuration.items():
            setattr(self, key, value)

    @property
    def is_valid(self) -> bool:
        """Check if the configuration has all required fields."""
        result: bool = is_valid_config(self.__dict__)[0]
        return result

    @property
    def log_file(self) -> str:
        """Get file name from logging configuration, if a file based log handler is defined."""
        logging = getattr(self, "logging", {})
        handlers = logging.get("handlers", {})
        file_handler = handlers.get("file", {})
        if file_handler:
            return str(file_handler["filename"])
        else:
            return ""

    def save(self, path: PathStr = "") -> None:
        """Saves configurations to the JSON file defined by `path` or :py:attr:`configuration_file`.

        If `path` is set, it will take presence and :attr:`.configuration_file` will be set accordingly.

        Args:
            path (PathStr): Full path of the JSON file to save to. If not specified, it will attempt to use the environment variable TIMESERIES_CONFIG before falling back to the default location `$HOME/.config/ssb_timeseries/timeseries_config.json`.

        Raises:
            ValueError: If `path` is not provided and :attr:`configuration_file` is not set.
        """
        from .io import fs

        if path:
            self.configuration_file = str(path)
        elif not self.configuration_file:
            raise ValueError(
                "Configuration file must have a value or path must be specified."
            )
        else:
            path = self.configuration_file

        fs.write_text(content=str(self), path=str(path), file_format="json")
        if not fs.exists(self.log_file):
            fs.touch(self.log_file)

        active_file(str(path))

    def __getitem__(self, item: str) -> Any | None:
        """Get the value of a configuration."""
        return getattr(self, str(item), None)

    def __eq__(self, other: Self | dict) -> bool:
        """Equality test."""
        if isinstance(other, dict):
            return self.__dict__ == other
        else:
            return self.__dict__ == other.__dict__

    def __str__(self) -> str:
        """Return timeseries configurations as JSON string."""
        return json.dumps(self.__dict__, sort_keys=True, indent=2)

    @classmethod
    def active(cls) -> Self:
        """Force reload the file identified by :py:const:`ENV_VAR_NAME` and return the configuration."""
        return cls(configuration_file=active_file())


class MissingEnvironmentVariableError(Exception):
    """The environment variable TIMESEREIS_CONFIG must be defined."""

    ...


class ValidationError(Exception):
    """Configuration validation error."""

    ...


def load_json_file(path: PathStr, error_on_missing: bool = False) -> dict:
    """Read configurations from a JSON file into a Config object."""
    from .io import fs

    if fs.exists(path):
        from_json = fs.read_json(path)
        if not isinstance(from_json, dict):
            from_json = json.loads(from_json)

        return from_json

    elif error_on_missing:
        raise FileNotFoundError(
            f"A configuration file at {path} file was specified, but does not exist."
        )
    else:
        return {}


class DictObject(object):  # noqa
    """Helper class to convert dict to object."""

    def __init__(self, dict_: dict) -> None:  # noqa: D107
        self.__dict__.update(dict_)

    @classmethod
    def from_dict(cls, d: dict):  # noqa: ANN206, D102
        return json.loads(json.dumps(d), object_hook=DictObject)


def presets(named_config: str) -> dict | ConfigDict:  # noqa: RUF100
    """Set configurations to predefined defaults.

    Raises:
        ValueError: If args is not 'home' | 'daplalab'.
    """
    if named_config in PRESETS:
        cfg = PRESETS[named_config]
        cfg["logging"]["handlers"]["file"]["filename"] = cfg.pop("log_file", "")
        return cfg
    else:
        raise ValueError(
            f"Named configuration preset '{named_config}' was not recognized."
        )


def main(*args: str | PathStr) -> None:
    """Set configurations to predefined defaults when run from command line.

    Use:
        ```
        poetry run timeseries-config <option>
        ```
    or
        ```
        python ./config.py <option>`
        ```

    Args:
        *args (str): 'home' | 'gcs' | 'daplalab'.

    Raises:
        ValueError: If args is not 'home' | 'gcs' | 'daplalab'. # noqa: DAR402

    """
    if args:
        config_identifier: PathStr = args[0]
    else:
        config_identifier = sys.argv[1]

    cfg = Config(preset=config_identifier)
    cfg.save(path=cfg.configuration_file)

    _config_logger.debug(
        f"Preset configuration '{config_identifier}' was applied:\n\t{cfg.__dict__}\nSaved to file: {cfg.configuration_file}.\nEnvironment variable set: {os.getenv('TIMESERIES_CONFIG')=}"
    )


def path_str(*args) -> str:
    """Concatenate paths as string: str(Path(...))."""
    return str(Path(*args))


if __name__ == "__main__":
    """Execute when called directly, ie not via import statements."""
    # ??? `poetry run timeseries-config <option>` does not appear to go this route.
    # --> not obvious that this is a good idea.
    print(f"Name of the script      : {sys.argv[0]=}")
    print(f"Arguments of the script : {sys.argv[1:]=}")
    main(sys.argv[1])
else:
    from .io import fs

    if active_file():
        if fs.exists(active_file()):
            CONFIGFILE = active_file()
        elif DAPLA_TEAM_CONTEXT:
            raise MissingEnvironmentVariableError(
                f"Environment variable {ENV_VAR_NAME} must be defined and point to a configuration file."
            )
        else:
            _config_logger.warning(
                f"No configuration file was found at {active_file()}.\nOther locations may be tried. Files found will be copied to the default location and the first candidate will be set to active, ie copied once more to {DEFAULTS['configuration_file']}"
            )
            if not fs.exists(CONFIGFILE):
                raise FileNotFoundError(
                    f"No configuration file was found at {active_file()}."
                )
    else:
        CONFIGFILE = ""  # PRESETS["defaults"]["configuration_file"]

    active_file(CONFIGFILE)
    CONFIG = Config(configuration_file=CONFIGFILE)
    """A Config object."""
    fs.touch(CONFIG.log_file)
    # do not save
    # CONFIG.save()



"""Provides helper functions for working with JSON data."""

from __future__ import annotations

import json
from datetime import datetime
from typing import Any
from typing import cast

from ..meta import TagDict
from ..types import SeriesType
from ..types import Temporality
from ..types import Versioning

# mypy: disable-error-code="assignment, union-attr"


def sanitize_for_json(d: dict) -> dict[str, Any]:
    """Recursively convert custom ssb-timeseries types to JSON-serializable strings."""
    if not isinstance(d, dict):
        return d
    sanitized_dict = {}
    for k, v in d.items():
        if isinstance(v, SeriesType | Versioning | Temporality):
            sanitized_dict[k] = str(v)
        elif isinstance(v, datetime):
            sanitized_dict[k] = v.isoformat()
        elif isinstance(v, dict):
            sanitized_dict[k] = sanitize_for_json(v)
        elif isinstance(v, list):
            sanitized_dict[k] = [
                sanitize_for_json(item) if isinstance(item, dict) else item
                for item in v
            ]
        else:
            sanitized_dict[k] = v
    return sanitized_dict


def tags_to_json(x: TagDict) -> dict[str, bytes]:
    """Serialize a tag dictionary into a format suitable for Parquet metadata.

    See Also:
        https://arrow.apache.org/docs/python/generated/pyarrow.schema.html
    """
    sanitized_tags = sanitize_for_json(x)
    j = {"json": json.dumps(sanitized_tags).encode("utf8")}
    return j


def tags_from_json(
    dict_with_json_string: dict[str | bytes, str | bytes],
    byte_encoded: bool = True,
) -> TagDict:  # dict[str, Any]:
    """Deserialize a tag dictionary from the Parquet metadata format.

    This is the reverse of `tags_to_json`.
    """
    if byte_encoded:
        d = json.loads(dict_with_json_string[b"json"].decode())
    else:
        d = json.loads(dict_with_json_string["json"])
    return cast(TagDict, d)



"""Defines common type aliases for the project."""

from __future__ import annotations

from collections.abc import Callable
from os import PathLike
from typing import TypeAlias

PathStr: TypeAlias = str | PathLike[str]
F: TypeAlias = Callable

from enum import Enum
from itertools import product

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self  # noqa: UP035 #backport to 3.10

# mypy: disable-error-code="type-arg, override"


class SuperEnum(Enum):
    """Generic enumeration helper."""

    @classmethod
    def to_dict(cls) -> dict:
        """Returns a dictionary representation of the enum."""
        return {e.name: e.value for e in cls}

    @classmethod
    def keys(cls) -> list:
        """Returns a list of all the enum keys."""
        return [item.name for item in cls]

    @classmethod
    def values(cls) -> list:
        """Returns a list of all the enum values."""
        return [item.value[0] for item in cls]

    @classmethod
    def descriptions(cls) -> list:
        """Returns a list of descriptions for all enum values."""
        return [item.value[1] for item in cls]

    def __eq__(self, other: Self) -> bool:
        """Equality test."""
        return repr(self) == repr(other)

    def __repr__(self) -> str:
        """Machine readable string representation, ideally sufficient to recreate object."""
        return f"{self.__class__.__name__}.{self.name}"

    def __str__(self) -> str:
        """Human readable string representation of object."""
        return self.name


class Versioning(SuperEnum):
    """Versioning refers to how revisions of data are identified (named)."""

    NONE = 0
    """Version control only. Versions are not accessible through API."""
    AS_OF = 1
    """Version identified by dates allows date arithemetic ('equals' | 'greater than'  | 'smaller than' | 'between')."""
    NAMES = 2
    """Versions identified by free text names.

    This functionality is under consideration, hence preparepared in the type system, but with not all the way through IO and core modules."""

    @property
    def date_columns(self) -> list[str]:
        """Returns the date columns for versioning."""
        return ["as_of"] if self == Versioning.AS_OF else []


class Temporality(SuperEnum):
    """Temporality describes the time dimensionality of each data point; notably duration or lack thereof."""

    NONE = 0
    """No temporal dimension."""
    AT = 1
    """Single points in time, expressed with (exact) 'valid_at' dates."""
    FROM_TO = 2
    """Duration periods, expressed with 'valid_from' (inclusive) and 'valid_to' (exclusive) dates."""

    @property
    def date_columns(self) -> list[str]:
        """Returns the data columns of the temporality."""
        if self == Temporality.AT:
            return ["valid_at"]
        if self == Temporality.FROM_TO:
            return ["valid_from", "valid_to"]
        return []


class SeriesType:
    """SeriesTypes are defined by combinations of attributes that have technical implications for time series datasets.

    Notable examples are Versioning and Temporality, but a few more may be added later.
    """

    def __init__(
        self,
        *type_def: str,
        versioning: Versioning | None = None,
        temporality: Temporality | None = None,
    ) -> None:
        """SeriesType constructor.

        Args:
            *type_def: Keyword strings identifying versioning and/or temporality.
            versioning: The versioning of the series.
            temporality: The temporality of the series.

        """
        # is_sparse (bool): Whether the series is sparse. Defaults to False.
        # dtype: numeric
        #   (without further specification for float/int/decimal)
        # partial support: bool
        # consider later: text
        words = {str(t).lower() for t in type_def}

        # Initialize with defaults
        _versioning = Versioning.NONE
        _temporality = Temporality.FROM_TO

        # Handle high-level keywords first
        if "simple" in words:
            _versioning = Versioning.NONE
            _temporality = Temporality.AT
        elif "estimate" in words:
            _versioning = Versioning.AS_OF
            _temporality = Temporality.AT
        else:
            # Fallback for other versioning keywords if no high-level keyword was found
            if {"asof", "as_of"} & words:
                _versioning = Versioning.AS_OF
            elif {"names"} & words:
                _versioning = Versioning.NAMES

            # Fallback for other temporality keywords if no high-level keyword was found
            if {"at"} & words:
                _temporality = Temporality.AT

        # Explicit arguments always override
        self.versioning = versioning if versioning is not None else _versioning
        self.temporality = temporality if temporality is not None else _temporality

    @classmethod
    def none_at(cls) -> Self:
        """Same as SeriesType.simple(): Shorthand for SeriesType(versioning=Versioning.NONE, temporality=Temporality.AT)."""
        return cls(versioning=Versioning.NONE, temporality=Temporality.AT)

    @classmethod
    def simple(cls) -> Self:
        """Same as SeriesType.none_at(): Shorthand for SeriesType(versioning=Versioning.NONE, temporality=Temporality.AT)."""
        return cls.none_at()

    @classmethod
    def from_to(cls) -> Self:
        """Shorthand for SeriesType(versioning=Versioning.NONE, temporality=Temporality.FROM_TO)."""
        return cls(versioning=Versioning.NONE, temporality=Temporality.FROM_TO)

    @classmethod
    def as_of_at(cls) -> Self:
        """Same as SeriesType.estimate(): Shorthand for SeriesType(versioning=Versioning.AS_OF, temporality=Temporality.AT)."""
        return cls(versioning=Versioning.AS_OF, temporality=Temporality.AT)

    @classmethod
    def as_of_from_to(cls) -> Self:
        """Shorthand for SeriesType(versioning=Versioning.AS_OF, temporality=Temporality.FROM_TO)."""
        return cls(versioning=Versioning.AS_OF, temporality=Temporality.FROM_TO)

    @classmethod
    def estimate(cls) -> Self:
        """Same as SeriesType.as_of_at(): Shorthand for SeriesType(versioning=Versioning.AS_OF, temporality=Temporality.AT)."""
        return cls.as_of_at()

    @classmethod
    def permutations(cls) -> list[str]:
        """Helper; returns ['<versioning>_<temporality>', ...] ."""
        return ["_".join(c) for c in product(Versioning.keys(), Temporality.keys())]

    @property
    def date_columns(self) -> list[str]:
        """Returns the data columns corresponding to the series type temporality."""
        return list(
            set(self.versioning.date_columns) | set(self.temporality.date_columns)
        )

    def __str__(self) -> str:
        """Helper; returns '<versioning>_<temporality>'."""
        return f"{self.versioning!s}_{self.temporality!s}"

    def __repr__(self) -> str:
        """Helper, returns code with required parameters to initialise given SeriesType."""
        return f"SeriesType({self.versioning!r},{self.temporality!r})"

    def __eq__(self, other: Self | None) -> bool:
        """Equality test."""
        if other is None:
            return False
        else:
            return repr(self) == repr(other)


def estimate_types() -> list[str]:
    """Helper; returns list of SeriesTypes for which Versioning is not NONE."""
    return ["_".join(c) for c in product(["AS_OF"], Temporality.keys())]


def seriestype_from_str(dir_name: str) -> SeriesType:
    """Helper; returns SeriesType from directory name."""
    match dir_name.lower():
        case "none_at":
            return SeriesType.none_at()
        case "simple":
            return SeriesType.simple()
        case "none_from_to":
            return SeriesType.from_to()
        case "as_of_at":
            return SeriesType.as_of_at()
        case "as_of_from_to":
            return SeriesType.as_of_from_to()
        case "estimate":
            return SeriesType.estimate()
        case _:
            raise ValueError(f"Invalid dir_name: {dir_name}")
